@@ -7,11 +7,6 @@ import struct
 import sys
 from contextlib import suppress
 
-if sys.version_info < (3, 8):
-    from typing_extensions import Final
-else:
-    from typing import Final
-
 import mbedtls._ringbuf as _rb
 
 from ._tls import (
@@ -19,6 +14,9 @@ from ._tls import (
     DTLSVersion,
     HandshakeStep,
     HelloVerifyRequest,
+)
+from ._tls import MbedTLSBuffer as TLSWrappedBuffer
+from ._tls import (
     NextProtocol,
     Purpose,
     RaggedEOF,
@@ -53,9 +51,6 @@ __all__ = (
     "ciphers_available",
 )
 
-TLS_BUFFER_CAPACITY: Final = 2 << 14
-# 32K (MBEDTLS_SSL_DTLS_MAX_BUFFERING)
-
 
 class TLSRecordHeader:
     """Encode/decode TLS record protocol format."""
@@ -149,8 +144,7 @@ class ClientContext(_BaseContext):
     def wrap_buffers(self, server_hostname):
         """Create an in-memory stream for TLS."""
         # PEP 543
-        self._set_hostname(server_hostname)
-        return TLSWrappedBuffer(self)
+        return TLSWrappedBuffer(self, server_hostname)
 
 
 class ServerContext(_BaseContext):
@@ -170,109 +164,13 @@ class ServerContext(_BaseContext):
         return TLSWrappedBuffer(self)
 
 
-class TLSWrappedBuffer:
-    # _pep543.TLSWrappedBuffer
-    def __init__(self, context):
-        self._output_buffer = _rb.RingBuffer(TLS_BUFFER_CAPACITY)
-        self._input_buffer = _rb.RingBuffer(TLS_BUFFER_CAPACITY)
-        context.set_bio(self._output_buffer, self._input_buffer)
-        self._context = context
-
-    def __repr__(self):
-        return "%s(%r)" % (type(self).__name__, self.context)
-
-    def __getstate__(self):
-        # We could make this pickable by copying the buffers.
-        raise TypeError(f"cannot pickle {self.__class__.__name__!r} object")
-
-    def read(self, amt):
-        # PEP 543
-        if amt <= 0:
-            return b""
-        buffer = bytearray(amt)
-        view = memoryview(buffer)
-        nread = 0
-        while nread != amt and not self._input_buffer.empty():
-            nread += self.readinto(view[nread:], amt - nread)
-        return bytes(buffer[:nread])
-
-    def readinto(self, buffer, amt):
-        # PEP 543
-        return self.context._readinto(buffer, amt)
-
-    def write(self, buffer):
-        # PEP 543
-        amt = self.context._write(buffer)
-        assert amt == len(buffer)
-        return len(self._output_buffer)
-
-    def do_handshake(self):
-        # PEP 543
-        self.context._do_handshake_step()
-
-    def _do_handshake_blocking(self, sock):
-        while self._context._state is not HandshakeStep.HANDSHAKE_OVER:
-            try:
-                self.context._do_handshake_step()
-                amt = sock.send(self.peek_outgoing(1024))
-                self.consume_outgoing(amt)
-            except WantReadError:
-                amt = sock.send(self.peek_outgoing(1024))
-                self.consume_outgoing(amt)
-            except WantWriteError:
-                data = sock.recv(1024)
-                self.receive_from_network(data)
-
-    def cipher(self):
-        # PEP 543
-        cipher = self.context._cipher()
-        if cipher is None:
-            return cipher
-        else:
-            return cipher[0]
-
-    def negotiated_protocol(self):
-        # PEP 543
-        return self.context._negotiated_protocol()
-
-    @property
-    def context(self):
-        # PEP 543
-        """The ``Context`` object this buffer is tied to."""
-        return self._context
-
-    def negotiated_tls_version(self):
-        # PEP 543
-        return self.context._negotiated_tls_version()
-
-    def shutdown(self):
-        # PEP 543
-        self.context._shutdown()
-
-    def receive_from_network(self, data):
-        # PEP 543
-        # Append data to input buffer.
-        self._input_buffer.write(data, len(data))
-
-    def peek_outgoing(self, amt):
-        # PEP 543
-        # Read from output buffer.
-        if amt == 0:
-            return b""
-        return self._output_buffer.peek(amt)
-
-    def consume_outgoing(self, amt):
-        """Consume `amt` bytes from the output buffer."""
-        # PEP 543
-        self._output_buffer.consume(amt)
-
-
 class TLSWrappedSocket:
     # _pep543.TLSWrappedSocket
     def __init__(self, socket, buffer):
         super().__init__()
         self._socket = socket
         self._buffer = buffer
+        self._context = buffer.context
         self._closed = False
 
     def __getstate__(self):
@@ -288,6 +186,10 @@ class TLSWrappedSocket:
     def __str__(self):
         return str(self._socket)
 
+    @property
+    def _handshake_state(self):
+        return self._buffer._handshake_state
+
     # PEP 543 requires the full socket API.
 
     @property
@@ -435,10 +337,19 @@ class TLSWrappedSocket:
     # PEP 543 adds the following methods.
 
     def do_handshake(self):
-        self._buffer._do_handshake_blocking(self._socket)
+        while self._handshake_state is not HandshakeStep.HANDSHAKE_OVER:
+            try:
+                self._buffer.do_handshake()
+            except WantReadError:
+                data = self._socket.recv(1024)
+                self._buffer.receive_from_network(data)
+            except WantWriteError:
+                in_transit = self._buffer.peek_outgoing(1024)
+                amt = self._socket.send(in_transit)
+                self._buffer.consume_outgoing(amt)
 
     def setcookieparam(self, param):
-        self.context._setcookieparam(param)
+        self._buffer.setcookieparam(param)
 
     def cipher(self):
         return self._buffer.cipher()
@@ -455,7 +366,4 @@ class TLSWrappedSocket:
 
     def unwrap(self):
         self._buffer.shutdown()
-        with suppress(OSError):
-            # shutdown may raise if the socket is not connected.
-            self._socket.shutdown(_socket.SHUT_RDWR)
         return self._socket
