@@ -1,6 +1,7 @@
 import datetime as dt
 import pickle
 import socket
+from contextlib import suppress
 
 import pytest
 
@@ -15,7 +16,16 @@ from mbedtls.x509 import CRT, CSR, BasicConstraints
 
 class TestPickle:
     @pytest.mark.parametrize(
-        "obj", (TLSConfiguration(), DTLSConfiguration()), ids=type
+        "obj",
+        (
+            TLSConfiguration(),
+            DTLSConfiguration(),
+            ClientContext(TLSConfiguration()),
+            ClientContext(DTLSConfiguration()),
+            ServerContext(TLSConfiguration()),
+            ServerContext(DTLSConfiguration()),
+        ),
+        ids=type,
     )
     def test_picklable(self, obj):
         assert obj == pickle.loads(pickle.dumps(obj))
@@ -23,10 +33,6 @@ class TestPickle:
     @pytest.mark.parametrize(
         "obj",
         (
-            ClientContext(TLSConfiguration()),
-            ClientContext(DTLSConfiguration()),
-            ServerContext(TLSConfiguration()),
-            ServerContext(DTLSConfiguration()),
             TLSSession(),
             TLSWrappedBuffer(ClientContext(DTLSConfiguration())),
             TLSWrappedBuffer(ClientContext(TLSConfiguration())),
@@ -436,26 +442,6 @@ class TestBaseContext:
         assert conf
         assert context.configuration is conf
 
-    def test_selected_npn_protocol(self, context):
-        assert context._selected_npn_protocol() is None
-
-    def test_cipher(self, context):
-        assert context._cipher() is None
-
-    def test_get_channel_binding(self, context):
-        assert context._get_channel_binding() is None
-
-    # def test_negotiated_tls_version(self, context):
-    #     assert context._negotiated_tls_version() is TLSVersion.SSLv3
-
-    @pytest.fixture
-    def tls_wrapped_buffer(self, context):
-        return TLSWrappedBuffer(context)
-
-    @pytest.mark.parametrize("repr_", (repr, str), ids=lambda f: f.__name__)
-    def test_repr_tls_wrapped_buffer(self, repr_, tls_wrapped_buffer):
-        assert isinstance(repr_(tls_wrapped_buffer), str)
-
 
 class TestClientContext(TestBaseContext):
     @pytest.fixture(params=[None, "hostname", "localhost"])
@@ -470,8 +456,8 @@ class TestClientContext(TestBaseContext):
         assert isinstance(context, ClientContext)
 
     def test_hostname(self, context, hostname):
-        _ = context.wrap_buffers(hostname)
-        assert context._hostname == hostname
+        wrapped_buffers = context.wrap_buffers(hostname)
+        assert wrapped_buffers._server_hostname == hostname
 
     def test_wrap_buffers(self, context):
         assert isinstance(context.wrap_buffers(None), TLSWrappedBuffer)
@@ -489,7 +475,7 @@ class TestServerContext(TestBaseContext):
         assert isinstance(context.wrap_buffers(), TLSWrappedBuffer)
 
 
-class TestTLSWrappedBuffer:
+class TestTLSHandshake:
     @pytest.fixture
     def server_hostname(self):
         return "hostname"
@@ -511,16 +497,16 @@ class TestTLSWrappedBuffer:
         )
 
     @pytest.fixture
-    def client(Self, cli_conf, server_hostname):
-        ctx = ClientContext(cli_conf).wrap_buffers(server_hostname)
-        yield ctx
-        ctx.shutdown()
+    def client(self, cli_conf, server_hostname):
+        buf = ClientContext(cli_conf).wrap_buffers(server_hostname)
+        yield buf
+        buf.shutdown()
 
     @pytest.fixture
     def server(self, srv_conf, server_hostname):
-        ctx = ServerContext(srv_conf).wrap_buffers()
-        yield ctx
-        ctx.shutdown()
+        buf = ServerContext(srv_conf).wrap_buffers()
+        yield buf
+        buf.shutdown()
 
     def test_e2e_handshake_and_communicate(self, client, server):
         def do_io(*, src, dst, amt=1024):
@@ -528,41 +514,49 @@ class TestTLSWrappedBuffer:
             src.consume_outgoing(len(in_transit))
             dst.receive_from_network(in_transit)
 
-        def do_handshake(*end_state_pair):
-            for end, state in end_state_pair:
+        def do_handshake(end, states):
+            while end._handshake_state is not states[0]:
+                # The backend goes through every state for both
+                # ends.  This is not relevant.
                 end.do_handshake()
-                assert end.context._state is state
+
+            for state in states:
+                assert end._handshake_state is state
+                if state is HandshakeStep.HANDSHAKE_OVER:
+                    break
+                with suppress(WantReadError, WantWriteError):
+                    end.do_handshake()
 
         do_handshake(
-            (server, HandshakeStep.CLIENT_HELLO),
-            (client, HandshakeStep.CLIENT_HELLO),
-            (client, HandshakeStep.SERVER_HELLO),
+            client,
+            (
+                HandshakeStep.HELLO_REQUEST,
+                HandshakeStep.CLIENT_HELLO,
+            ),
         )
 
         do_io(src=client, dst=server)
         do_handshake(
-            (server, HandshakeStep.SERVER_HELLO),
-            (server, HandshakeStep.SERVER_CERTIFICATE),
-            (server, HandshakeStep.SERVER_KEY_EXCHANGE),
-            (server, HandshakeStep.CERTIFICATE_REQUEST),
-            (server, HandshakeStep.SERVER_HELLO_DONE),
-            (server, HandshakeStep.CLIENT_CERTIFICATE),
-            (server, HandshakeStep.CLIENT_KEY_EXCHANGE),
+            server,
+            (
+                HandshakeStep.SERVER_HELLO,
+                HandshakeStep.SERVER_CERTIFICATE,
+                HandshakeStep.SERVER_KEY_EXCHANGE,
+                HandshakeStep.CERTIFICATE_REQUEST,
+                HandshakeStep.SERVER_HELLO_DONE,
+            ),
         )
         assert client.negotiated_protocol() == server.negotiated_protocol()
 
         do_io(src=server, dst=client)
         do_handshake(
-            (client, HandshakeStep.SERVER_CERTIFICATE),
-            (client, HandshakeStep.SERVER_KEY_EXCHANGE),
-            (client, HandshakeStep.CERTIFICATE_REQUEST),
-            (client, HandshakeStep.SERVER_HELLO_DONE),
-            (client, HandshakeStep.CLIENT_CERTIFICATE),
-            (client, HandshakeStep.CLIENT_KEY_EXCHANGE),
-            (client, HandshakeStep.CERTIFICATE_VERIFY),
-            (client, HandshakeStep.CLIENT_CHANGE_CIPHER_SPEC),
-            (client, HandshakeStep.CLIENT_FINISHED),
-            (client, HandshakeStep.SERVER_CHANGE_CIPHER_SPEC),
+            client,
+            (
+                HandshakeStep.CLIENT_KEY_EXCHANGE,
+                HandshakeStep.CERTIFICATE_VERIFY,
+                HandshakeStep.CLIENT_CHANGE_CIPHER_SPEC,
+                HandshakeStep.CLIENT_FINISHED,
+            ),
         )
         assert (
             client.negotiated_tls_version() == server.negotiated_tls_version()
@@ -570,23 +564,154 @@ class TestTLSWrappedBuffer:
 
         do_io(src=client, dst=server)
         do_handshake(
-            (server, HandshakeStep.CERTIFICATE_VERIFY),
-            (server, HandshakeStep.CLIENT_CHANGE_CIPHER_SPEC),
-            (server, HandshakeStep.CLIENT_FINISHED),
-            (server, HandshakeStep.SERVER_CHANGE_CIPHER_SPEC),
-            (server, HandshakeStep.SERVER_FINISHED),
-            (server, HandshakeStep.FLUSH_BUFFERS),
-            (server, HandshakeStep.HANDSHAKE_WRAPUP),
-            (server, HandshakeStep.HANDSHAKE_OVER),
+            server,
+            (
+                HandshakeStep.SERVER_CHANGE_CIPHER_SPEC,
+                HandshakeStep.SERVER_FINISHED,
+                HandshakeStep.FLUSH_BUFFERS,
+                HandshakeStep.HANDSHAKE_WRAPUP,
+                HandshakeStep.HANDSHAKE_OVER,
+            ),
         )
 
         do_io(src=server, dst=client)
+        do_handshake(client, (HandshakeStep.HANDSHAKE_OVER,))
+        assert client.cipher() == server.cipher()
+
+        secret = b"a very secret message"
+
+        amt = client.write(secret)
+        do_io(src=client, dst=server)
+        assert server.read(amt) == secret
+
+        amt = server.write(secret)
+        do_io(src=server, dst=client)
+        assert client.read(amt) == secret
+
+
+class TestDTLSHandshake:
+    @pytest.fixture
+    def server_hostname(self):
+        return "hostname"
+
+    @pytest.fixture
+    def psk(self):
+        return ("cli", b"secret")
+
+    @pytest.fixture
+    def cli_conf(self, psk):
+        return DTLSConfiguration(
+            pre_shared_key=psk, validate_certificates=False
+        )
+
+    @pytest.fixture
+    def srv_conf(self, psk):
+        return DTLSConfiguration(
+            pre_shared_key_store=dict((psk,)), validate_certificates=False
+        )
+
+    @pytest.fixture
+    def client(self, cli_conf, server_hostname):
+        buf = ClientContext(cli_conf).wrap_buffers(server_hostname)
+        yield buf
+        buf.shutdown()
+
+    @pytest.fixture
+    def server(sefl, srv_conf):
+        buf = ServerContext(srv_conf).wrap_buffers()
+        yield buf
+        buf.shutdown()
+
+    @pytest.fixture
+    def cli_ctx(Self, cli_conf):
+        return ClientContext(cli_conf)
+
+    @pytest.fixture
+    def srv_ctx(self, srv_conf):
+        return ServerContext(srv_conf)
+
+    def test_e2e_handshake_and_communicate(self, client, server):
+        def do_io(*, src, dst, amt=1024):
+            in_transit = src.peek_outgoing(amt)
+            src.consume_outgoing(len(in_transit))
+            dst.receive_from_network(in_transit)
+
+        def do_handshake(end, states):
+            while end._handshake_state is not states[0]:
+                # The backend goes through every state for both
+                # ends.  This is not relevant.
+                end.do_handshake()
+
+            for state in states:
+                assert end._handshake_state is state
+                if state is HandshakeStep.HANDSHAKE_OVER:
+                    break
+                with suppress(WantReadError, WantWriteError):
+                    end.do_handshake()
+
+        do_handshake(
+            client,
+            (
+                HandshakeStep.HELLO_REQUEST,
+                HandshakeStep.CLIENT_HELLO,
+            ),
+        )
+
+        do_io(src=client, dst=server)
+        server.setcookieparam("🍪🍪🍪".encode("utf-8"))
+        with pytest.raises(HelloVerifyRequest):
+            do_handshake(
+                server,
+                (
+                    HandshakeStep.SERVER_HELLO,
+                    HandshakeStep.SERVER_HELLO_VERIFY_REQUEST_SENT,
+                ),
+            )
+
+        do_handshake(server, (HandshakeStep.HELLO_REQUEST,))
+        do_io(src=server, dst=client)
+        do_handshake(client, (HandshakeStep.CLIENT_HELLO,))
+        do_io(src=client, dst=server)
+        server.setcookieparam("🍪🍪🍪".encode("utf-8"))
+        do_handshake(
+            server,
+            (
+                HandshakeStep.SERVER_HELLO,
+                HandshakeStep.SERVER_CERTIFICATE,
+                HandshakeStep.SERVER_KEY_EXCHANGE,
+                HandshakeStep.CERTIFICATE_REQUEST,
+                HandshakeStep.SERVER_HELLO_DONE,
+            ),
+        )
+        assert client.negotiated_protocol() == server.negotiated_protocol()
+
+        do_io(src=server, dst=client)
+        do_handshake(
+            client,
+            (
+                HandshakeStep.CLIENT_KEY_EXCHANGE,
+                HandshakeStep.CERTIFICATE_VERIFY,
+                HandshakeStep.CLIENT_CHANGE_CIPHER_SPEC,
+                HandshakeStep.CLIENT_FINISHED,
+            ),
+        )
+        assert (
+            client.negotiated_tls_version() == server.negotiated_tls_version()
+        )
+
+        do_io(src=client, dst=server)
         do_handshake(
-            (client, HandshakeStep.SERVER_FINISHED),
-            (client, HandshakeStep.FLUSH_BUFFERS),
-            (client, HandshakeStep.HANDSHAKE_WRAPUP),
-            (client, HandshakeStep.HANDSHAKE_OVER),
+            server,
+            (
+                HandshakeStep.SERVER_CHANGE_CIPHER_SPEC,
+                HandshakeStep.SERVER_FINISHED,
+                HandshakeStep.FLUSH_BUFFERS,
+                HandshakeStep.HANDSHAKE_WRAPUP,
+                HandshakeStep.HANDSHAKE_OVER,
+            ),
         )
+        do_io(src=server, dst=client)
+        do_handshake(client, (HandshakeStep.HANDSHAKE_OVER,))
         assert client.cipher() == server.cipher()
 
         secret = b"a very secret message"
